@@ -3,12 +3,17 @@
 This class separates out the parsing of different message types from the EmBody device's file format. The first part
 consists of all the different messages wrapped as subclasses of the ProtocolMessage dataclass. The bottom part
 provides access methods for parsing one and one message from a bytes object.
+
+This module uses a dictionary-based registry pattern (_FILE_MESSAGE_REGISTRY) for O(1) message type lookups.
+Temperature conversions use the TEMPERATURE_SCALE_FACTOR constant (0.0078125 = 1/128).
 """
 
 import struct
 from dataclasses import dataclass
-from dataclasses import field
-from typing import Optional
+
+# Temperature sensor conversion factor (degrees Celsius per raw unit)
+# This factor converts raw sensor values to degrees Celsius
+TEMPERATURE_SCALE_FACTOR = 0.0078125  # 1/128
 
 
 @dataclass
@@ -17,19 +22,17 @@ class ProtocolMessage:
     unpack_format = ""
 
     @classmethod
-    def default_length(cls, version: Optional[tuple[int, int, int]] = None) -> int:
+    def default_length(cls, version: tuple[int, int, int] | None = None) -> int:
         return struct.calcsize(cls.unpack_format)
 
-    def length(self, version: Optional[tuple[int, int, int]] = None) -> int:
+    def length(self, version: tuple[int, int, int] | None = None) -> int:
         return self.__class__.default_length(version)
 
     @classmethod
-    def decode(cls, data: bytes, version: Optional[tuple[int, int, int]] = None):
+    def decode(cls, data: bytes, version: tuple[int, int, int] | None = None):
         if len(data) < cls.default_length(version):
             raise BufferError("Buffer too short for message")
-        return cls(
-            *(struct.unpack(cls.unpack_format, data[0 : cls.default_length(version)]))
-        )
+        return cls(*(struct.unpack(cls.unpack_format, data[0 : cls.default_length(version)])))
 
 
 @dataclass
@@ -37,7 +40,7 @@ class TimetickedMessage(ProtocolMessage):
     two_lsb_of_timestamp = None  # Dataclass workaround. Not specified with type to avoid having it as a dataclass field
 
     @classmethod
-    def decode(cls, data: bytes, version: Optional[tuple[int, int, int]] = None):
+    def decode(cls, data: bytes, version: tuple[int, int, int] | None = None):
         if len(data) < cls.default_length(version):
             raise BufferError("Buffer too short for message")
         tuples = struct.unpack(cls.unpack_format, data[0 : cls.default_length(version)])
@@ -110,11 +113,11 @@ class PpgRaw(TimetickedMessage):
     ppg: int
 
     @classmethod
-    def default_length(cls, version: Optional[tuple[int, int, int]] = None) -> int:
+    def default_length(cls, version: tuple[int, int, int] | None = None) -> int:
         return 8
 
     @classmethod
-    def decode(cls, data: bytes, version: Optional[tuple[int, int, int]] = None):
+    def decode(cls, data: bytes, version: tuple[int, int, int] | None = None):
         if len(data) < cls.default_length(version):
             raise BufferError("Buffer too short for message")
         ts_lsb = int.from_bytes(data[0:2], byteorder="big", signed=False)
@@ -133,11 +136,11 @@ class PpgRawAll(TimetickedMessage):
     ppg_ir: int
 
     @classmethod
-    def default_length(cls, version: Optional[tuple[int, int, int]] = None) -> int:
+    def default_length(cls, version: tuple[int, int, int] | None = None) -> int:
         return 11
 
     @classmethod
-    def decode(cls, data: bytes, version: Optional[tuple[int, int, int]] = None):
+    def decode(cls, data: bytes, version: tuple[int, int, int] | None = None):
         if len(data) < cls.default_length(version):
             raise BufferError("Buffer too short for message")
         ts_lsb = int.from_bytes(data[0:2], byteorder="big", signed=False)
@@ -225,7 +228,7 @@ class Temperature(TimetickedMessage):
     temp_raw: int
 
     def temp_celsius(self) -> float:
-        return self.temp_raw * 0.0078125
+        return self.temp_raw * TEMPERATURE_SCALE_FACTOR
 
 
 @dataclass
@@ -238,49 +241,33 @@ class PulseRawList(TimetickedMessage):
     len: int = 6  # actual length, since this is instance specific, not static
 
     @classmethod
-    def default_length(cls, version: Optional[tuple[int, int, int]] = None) -> int:
+    def default_length(cls, version: tuple[int, int, int] | None = None) -> int:
         """Return a dummy value, since this is instance specific for this class."""
         return 6
 
-    def length(self, version: Optional[tuple[int, int, int]] = None) -> int:
+    def length(self, version: tuple[int, int, int] | None = None) -> int:
         return self.len
 
     @classmethod
-    def decode(cls, data: bytes, version: Optional[tuple[int, int, int]] = None):
+    def decode(cls, data: bytes, version: tuple[int, int, int] | None = None):
         if len(data) < 3:
-            raise BufferError(
-                f"Buffer too short for message. Received {len(data)} bytes, expected at least 10 bytes"
-            )
+            raise BufferError(f"Buffer too short for message. Received {len(data)} bytes, expected at least 10 bytes")
         (tick,) = struct.unpack("<H", data[0:2])
         (format_and_sizes,) = struct.unpack("<B", data[2:3])
-        fmt, no_of_ecgs, no_of_ppgs = PulseRawList.to_format_and_lengths(
-            format_and_sizes
-        )
+        fmt, no_of_ecgs, no_of_ppgs = PulseRawList.to_format_and_lengths(format_and_sizes)
         ecgs = []
         ppgs = []
-        bytes_per_ecg_and_ppg = (
-            1 if fmt == 0 else 2 if fmt == 1 else 3 if fmt == 2 else 4
-        )
-        length = (
-            3
-            + (no_of_ecgs * bytes_per_ecg_and_ppg)
-            + (no_of_ppgs * bytes_per_ecg_and_ppg)
-        )
+        bytes_per_ecg_and_ppg = 1 if fmt == 0 else 2 if fmt == 1 else 3 if fmt == 2 else 4
+        length = 3 + (no_of_ecgs * bytes_per_ecg_and_ppg) + (no_of_ppgs * bytes_per_ecg_and_ppg)
         if len(data) < length:
-            raise BufferError(
-                f"Buffer too short for message. Received {len(data)} bytes, expected {length} bytes"
-            )
+            raise BufferError(f"Buffer too short for message. Received {len(data)} bytes, expected {length} bytes")
         pos = 3
         for _ in range(no_of_ecgs):
-            ecg = int.from_bytes(
-                data[pos : pos + bytes_per_ecg_and_ppg], byteorder="little", signed=True
-            )
+            ecg = int.from_bytes(data[pos : pos + bytes_per_ecg_and_ppg], byteorder="little", signed=True)
             ecgs.append(ecg)
             pos += bytes_per_ecg_and_ppg
         for _ in range(no_of_ppgs):
-            ppg = int.from_bytes(
-                data[pos : pos + bytes_per_ecg_and_ppg], byteorder="little", signed=True
-            )
+            ppg = int.from_bytes(data[pos : pos + bytes_per_ecg_and_ppg], byteorder="little", signed=True)
             ppgs.append(ppg)
             pos += bytes_per_ecg_and_ppg
         msg = PulseRawList(
@@ -302,6 +289,102 @@ class PulseRawList(TimetickedMessage):
         return fmt, no_of_ecgs, no_of_ppgs
 
 
+@dataclass
+class PulseBlockEcg(TimetickedMessage):
+    time: int
+    channel: int
+    num_samples: int
+    samples: list[int]
+    pkg_length: int
+
+    @classmethod
+    def default_length(cls, version: tuple[int, int, int] | None = None) -> int:
+        """Return a dummy value, since this is instance specific for this class."""
+        return 14
+
+    def length(self, version: tuple[int, int, int] | None = None) -> int:
+        return self.pkg_length
+
+    @classmethod
+    def decode(cls, data: bytes, version: tuple[int, int, int] | None = None) -> "PulseBlockEcg":
+        if len(data) < 14:
+            raise BufferError(f"Buffer too short for message. Received {len(data)} bytes, expected at least 13 bytes")
+        channel = data[0]
+        packed_ecgs = data[1]
+        pkg_length = 1 + 1 + 8 + 4 + (packed_ecgs * 2)
+        if len(data) < pkg_length:
+            raise BufferError(f"Buffer too short for message. Received {len(data)} bytes, expected at least 13 bytes")
+        (time,) = struct.unpack("<Q", data[2:10])
+        samples = []
+        ref = int.from_bytes(data[10:14], byteorder="little", signed=True)
+        samples.append(ref)
+        pos = 14
+        for _ in range(packed_ecgs):
+            sample = ref + int.from_bytes(data[pos : pos + 2], byteorder="little", signed=True)
+            samples.append(sample)
+            pos += 2
+        msg = PulseBlockEcg(
+            time=time,
+            channel=channel,
+            num_samples=packed_ecgs + 1,
+            samples=samples,
+            pkg_length=pkg_length,
+        )
+        return msg
+
+    def encode(self) -> bytes:
+        payload = struct.pack("<H", 0)
+        return payload
+
+
+@dataclass
+class PulseBlockPpg(TimetickedMessage):
+    time: int
+    channel: int
+    num_samples: int
+    samples: list[int]
+    pkg_length: int
+
+    @classmethod
+    def default_length(cls, version: tuple[int, int, int] | None = None) -> int:
+        """Return a dummy value, since this is instance specific for this class."""
+        return 14
+
+    def length(self, version: tuple[int, int, int] | None = None) -> int:
+        return self.pkg_length
+
+    @classmethod
+    def decode(cls, data: bytes, version: tuple[int, int, int] | None = None) -> "PulseBlockPpg":
+        if len(data) < 13:
+            raise BufferError(f"Buffer too short for message. Received {len(data)} bytes, expected at least 13 bytes")
+        channel = data[0]
+        packed_ppgs = data[1]
+        pkg_length = 1 + 1 + 8 + 4 + (packed_ppgs * 2)
+        if len(data) < pkg_length:
+            raise BufferError(f"Buffer too short for message. Received {len(data)} bytes, expected at least 13 bytes")
+        (time,) = struct.unpack("<Q", data[2:10])
+        samples = []
+        ref = int.from_bytes(data[10:14], byteorder="little", signed=True)
+        samples.append(ref)
+        pos = 14
+        for _ in range(packed_ppgs):
+            sample = ref + int.from_bytes(data[pos : pos + 2], byteorder="little", signed=True)
+            samples.append(sample)
+            pos += 2
+        msg = PulseBlockPpg(
+            time=time,
+            channel=channel,
+            num_samples=packed_ppgs + 1,
+            samples=samples,
+            pkg_length=pkg_length,
+        )
+        return msg
+
+    def encode(self) -> bytes:
+        payload = struct.pack("<H", 0)
+        return payload
+
+
 @dataclass
 class BatteryDiagnostics(TimetickedMessage):
     struct_format = "<IIHHhhHHHH"
@@ -318,11 +401,11 @@ class BatteryDiagnostics(TimetickedMessage):
     vfsoc: int  # % *100  (0-100.00 %) Voltage based fuelgauge State Of Charge
 
     @classmethod
-    def default_length(cls, version: Optional[tuple[int, int, int]] = None) -> int:
+    def default_length(cls, version: tuple[int, int, int] | None = None) -> int:
         return struct.calcsize(cls.struct_format)
 
     @classmethod
-    def decode(cls, data: bytes, version: Optional[tuple[int, int, int]] = None):
+    def decode(cls, data: bytes, version: tuple[int, int, int] | None = None):
         if len(data) < cls.default_length(version):
             raise BufferError("Buffer too short for message")
         ts_lsb = int.from_bytes(data[0:2], byteorder="little", signed=False)
@@ -336,53 +419,48 @@ class BatteryDiagnostics(TimetickedMessage):
         return msg
 
 
-def decode_message(
-    data: bytes, version: Optional[tuple[int, int, int]] = None
-) -> ProtocolMessage:
+# File protocol message registry
+_FILE_MESSAGE_REGISTRY: dict[int, type[ProtocolMessage]] = {
+    0x01: Header,
+    0x71: Timestamp,
+    0xAC: ImuRaw,
+    0xA4: Imu,
+    0xB1: PpgRaw,
+    0xA2: PpgRawAll,
+    0xA1: BatteryLevel,
+    0xA5: HeartRate,
+    0xAD: HeartRateInterval,
+    0x74: NoOfPpgValues,
+    0xA9: ChargeState,
+    0xAA: BeltOnBody,
+    0x07: AfeSettingsAll,
+    0xB2: AccRaw,
+    0xB3: GyroRaw,
+    0xB4: Temperature,
+    0xB6: PulseRawList,
+    0xB8: PulseBlockEcg,
+    0xB9: PulseBlockPpg,
+    0xBB: BatteryDiagnostics,
+}
+
+
+def decode_message(data: bytes, version: tuple[int, int, int] | None = None) -> ProtocolMessage:
     """Decodes a bytes object into proper subclass of ProtocolMessage.
 
     raises LookupError if unknown message type.
     """
-
     message_type = data[0]
-    if message_type == 0x01:
-        return Header.decode(data[1:], version)
-    elif message_type == 0x71:
-        return Timestamp.decode(data[1:], version)
-    elif message_type == 0xAC:
-        return ImuRaw.decode(data[1:], version)
-    elif message_type == 0xA4:
-        return Imu.decode(data[1:], version)
-    elif message_type == 0xB1:
-        return PpgRaw.decode(data[1:], version)
-    elif message_type == 0xA2:
-        return PpgRawAll.decode(data[1:], version)
-    elif message_type == 0xA1:
-        return BatteryLevel.decode(data[1:], version)
-    elif message_type == 0xA5:
-        return HeartRate.decode(data[1:], version)
-    elif message_type == 0xAD:
-        return HeartRateInterval.decode(data[1:], version)
-    elif message_type == 0x74:
-        return NoOfPpgValues.decode(data[1:], version)
-    elif message_type == 0xA9:
-        return ChargeState.decode(data[1:], version)
-    elif message_type == 0xAA:
-        return BeltOnBody.decode(data[1:], version)
-    elif message_type == 0x06 and isinstance(version, tuple) and version >= (4, 0, 1):
-        return AfeSettings.decode(data[1:], version)
-    elif message_type == 0x06:
-        return AfeSettingsOld.decode(data[1:], version)
-    elif message_type == 0x07:
-        return AfeSettingsAll.decode(data[1:], version)
-    elif message_type == 0xB2:
-        return AccRaw.decode(data[1:], version)
-    elif message_type == 0xB3:
-        return GyroRaw.decode(data[1:], version)
-    elif message_type == 0xB4:
-        return Temperature.decode(data[1:], version)
-    elif message_type == 0xB6:
-        return PulseRawList.decode(data[1:], version)
-    elif message_type == 0xBB:
-        return BatteryDiagnostics.decode(data[1:], version)
-    raise LookupError(f"Unknown message type {hex(message_type)}")
+
+    # Special handling for version-dependent AfeSettings
+    if message_type == 0x06:
+        if isinstance(version, tuple) and version >= (4, 0, 1):
+            return AfeSettings.decode(data[1:], version)
+        else:
+            return AfeSettingsOld.decode(data[1:], version)
+
+    # Lookup message class from registry
+    message_class = _FILE_MESSAGE_REGISTRY.get(message_type)
+    if message_class is None:
+        raise LookupError(f"Unknown message type {hex(message_type)}")
+
+    return message_class.decode(data[1:], version)
